/*
   Approssima righelli di Zanasi
   
   @TITLE Approz
   @AUTHOR R.Corda

   Copyright (c) 2001, Roberto Corda (roccopapaleo@yahoo.it)
   All rights reserved.

   Public Domain con nessuna garanzia.
*/

/*
 *
 * Approssima i righelli ottimali di zanasi (RoZ)
 * cerca i migliori RZ in maniera euristica 
 * 
 * approz [a|s][m|d] <number>
 * approz h 

 * d(ebug) : stampa valori intermedi
 * m(ultipla): ricerca sino a number 
 * null: ricerca il singolo numero 

 * a(dd): per aggiunte sucessive a partire da [0,1]
 * s(ub): per selezione a partire da [0..n]

 * h(elp): stampa help
*/

#define GROUP "ACME inc 2001"
#define TITLE "approz"
#define AUTHOR "Roberto Corda"
#define VERSION "v0.2.27052001"

#include <stdio.h>
#include <stdlib.h>

#define better(a,b) ((a.tac < b.tac) || (a.tac == b.tac) && (a.corr > b.corr))
#define coolvar(a,best,coeff) ((a.corr*coeff) > best.corr)
#define coollev(a,best,level) ((a.tac-level) < best.tac)

#define MAXTAG 20
  /* MAXN = [Tr(MAXTAG)] */
#define MAXN 102
#define MAXSTACK 1500
#define FALSE 0
#define TRUE 1
#define MIN 1
#define MAX 2

#define OVRFLW -1
#define EMPTY  -2

typedef struct {

 int x[MAXN];      /* tacca i-esima */
 int chkroz[MAXN]; /* quante volte appare la distanza i */
 int tac;          /* numero tacche rz */
 int sum;          /* somma dei chkroz */
 float corr;       /* varianza/media */
 float ave;        /* media */

} trighello;


trighello rstack[MAXSTACK];
int stack=0, debug = FALSE;
float coeff;

void swap(int a, int b, trighello *r);

int equal(trighello *a, trighello *b) {
 int i;
 if (a->tac != b->tac) return FALSE;

 for (i=0; i<a->tac; i++) {
  if (a->x[i] != b->x[i]) return FALSE;
 }

 return TRUE;
}


/* TRUE se t non in s */
int notinside(trighello *t, trighello *s, int stack) {
 int i;

 for (i=0; i<stack; i++){
  if (equal(t, s+i)) return FALSE;
 }

 return TRUE;
}

/* TRUE se a > b */
int gte(trighello *a, trighello *b) {
 int i;

 if (a->tac >= b->tac) return TRUE;

 return FALSE;
}

/* pone x[0] == 0 e sceglie la riflessione piu' piccola */
void normalizza(trighello *r) {
 int i, p, s, d, refl;

 refl = FALSE;

 /* sceglie la direzione che minimizza */
 /* procede da destra e da sinistra sino alla prima asimmetria */
 for (s=0, d=r->tac-1; s<r->tac/2-1; s++, d--) {
  if ( (r->x[s] - r->x[s+1] ) > (r->x[d-1] - r->x[d] )) {
     refl = TRUE;
  }
 }

 if (refl) {
  /* fa la riflessione */
  for (s=0, d=r->tac-1; s<r->tac/2-1; s++, d--) {
   swap(s,d,r) ; 
  }
 } 


 /* trasla in modo che x[0] == 0 */
 if (p = r->x[0]) 
   for (i=r->tac-1; i>=0; i-- )
      r->x[i] -= p;

}

/* ritorna la posizione del elemento piu' grande in stack */
int max(trighello *s, int stack) {
 int p = 0,i;

 for (i=0; i<stack; i++){
  if (gte(s+i, s+p)) p = i;
 }

 return p;
}

/* ritorna la posizione del elemento piu' piccolo in stack */
int min(trighello *s, int stack) {
 int p = 0,i;

 for (i=0; i<stack; i++){
  if (gte(s+p, s+i)) p = i;
 }

 return p;
}

/* mette l'elemento in pila solo se non gia' presente */
int pushr(trighello p, trighello *r, int *s) {
  if (*s<= MAXSTACK) {
   if ((*s == 0) || (notinside(&p, r, *s))) {
    r[*s] = p;
    (*s)++;
   }
  }
 else {
   fprintf(stderr, "Error: stack overflow\n");
   return OVRFLW;
  }
}

void swap(int a, int b, trighello *r) {
 trighello rswap;

 rswap = r[a];
 r[a] = r[b];
 r[b] = rswap;
}

/* restituisce l'indice dell'elemento in *s*/
/* se mode == MIN il migliore */
/* se mode == MAX il peggiore */
/* il primo in pila altrimenti */
/* EMPTY<0 se pila vuota */
int popr(trighello *r, int *s, int mode){
 int p;
 

 if (*s>0) {

  if (mode == MIN)
   p = min(r,stack);
  if (mode == MAX)
   p = max(r,stack);

  if ((mode == MIN) || (mode == MAX)){
   swap(p,*s-1,r) ;
  }

  return --(*s);
 } else {
   fprintf(stderr, "Error: stack empty\n");
   return EMPTY;
  }
}

/* calcola il campo chkroz del righello
   calcola quante volte compare ogni possibile distanza*/
void calcChkroz(trighello *r, int lev){
 int i;

 if (lev < r->tac) {
  for (i=lev; i<r->tac; i++){
   r->chkroz[i-lev]++;
   r->sum++;
  }
  calcChkroz(r, lev+1);
 }

}

/* calcola sino a quanto conta il righello r */
int calcRZ(trighello *r, int n) {
 int i;

 for (i=-1;
     ((i+1<=n) && ( r->chkroz[i+1]));
     i++);


 return i;
}

/* calcola varianza semplificata ovvero average(abs(x-ave)) */
float varianza(trighello r) {
 float ave = r.ave;
 float sum = r.sum - r.x[0];
 float var = 0;
 int i;
 for (i = 1; i<r.tac; i++) {
   var += (float) r.x[i] * (float) abs((float)r.x[i] - ave);
 }

 var /= sum;

 return var;
}

float media(trighello r) {
 float ave;
 int i;

 ave = 0;

 for (i=0; i<r.tac; i++) {
  ave += r.x[i] * r.x[i] ;
 }

 if (r.sum)
   return ave/ (float)r.sum;
 else {
  fprintf(stderr, "Error: sum 0 on average");
  exit(1);
 }
}

/* la sottrazione [-] */
/* FIX: non ci sono controlli */
trighello sub(trighello *rinput, int p) {
 int i,dist;
 trighello r;
 float sum, var, ave;

 r = *rinput;


 /* ricalcola chkroz e sum */
 for (i=0; i<r.tac; i++){

  dist = abs(r.x[p]- r.x[i]);
  r.sum--;
  r.chkroz[dist]--;
 }

 /* sposta */
 for (i=p; i<r.tac; i++) {
  r.x[i] = r.x[i+1];
 }

 --r.tac;
 /* calcola media */
 r.ave = media(r);

 /* calcola coefficente di correlazione var[x]/E[x] */
 r.corr = varianza(r)/ r.ave;

 return r;
}

/* mette nello stack l'insieme degli r [-] a */
void pushSub(trighello *r, int a){
 trighello result;
 int i;

 for (i=0; i<r->tac; i++){
  result = sub(r, i);
  if (calcRZ(&result,a) == a)
   pushr(result, rstack, &stack) ;
 }
}

/* riempie lo stack cont tutti gli elementi
   dell'insieme [0...a] [-] xi */
void init(int a) {
 int i;
 trighello rinit, result;

 /* [0,1,..., a] */
 for (i=0; i<=a; i++){
   rinit.x[i] = i;
   rinit.chkroz[i]=0;
 }
 rinit.tac = a+1;
 rinit.sum = 0;

 calcChkroz(&rinit,0);

 /* calcola media */
 rinit.ave = media(rinit);

 /* calcola correlazione*/
 rinit.corr = varianza(rinit)/ rinit.ave;

 pushSub(&rinit, a);
}

void printRZ(FILE *f, trighello r, int n){
 int i;


 fprintf(f, "rz(%2u): (t=%2u s=%u) [", n, r.tac,stack);
 for (i=0; i<r.tac-1; i++){
  fprintf(f, "%2u,", r.x[i]);

 }
 fprintf(f, "%2u] (corr = %3f ave = %3f)\n", r.x[r.tac-1] ,r.corr, r.ave);
 fflush(f);
}

/*
 * estrae dalla pila
 * sottrai gli xi
 * rimette in coda se sono ancora
 */
void approz(int n){

 trighello rcheck,best;
 int ibest, icheck;



 if ((ibest = popr(rstack, &stack, MIN)) == EMPTY)
   {
    fprintf(stderr, "Stack Empty \n");
    exit(0);
   }

 best = rstack[ibest];

 if (debug) printRZ(stderr, best,n);

 while (stack) {
  if (stack <(MAXSTACK/2))
    {

     icheck = popr(rstack, &stack, MAX);
     if (icheck != EMPTY)
       rcheck = rstack[icheck];
    }
  else
    {
     icheck = popr(rstack, &stack, MIN);
     if (icheck != EMPTY)
       rcheck = rstack[icheck];
    }

  if (icheck != EMPTY) {
   if (better(rcheck, best)) {
    best = rcheck;
    if (debug) printRZ(stderr, best,n);
    pushSub(&rcheck,n);
   } else if (coolvar(rcheck, best, coeff)) {
    pushSub(&rcheck,n);
   }
  }
 }
 printRZ(stdout, best,n);
}

void addroz(int n) {

 trighello rcheck,best;
 int ibest, icheck;

 /* metti in pila [0,1] */
 addinit();

 /* aggiungi tutte le somme di quelli in pila */
 /* normalizzate e senza doppioni */
 if ((ibest = popr(rstack, &stack, MIN)) == EMPTY)
   {
    fprintf(stderr, "Stack Empty \n");
    exit(0);
   }

 best = rstack[ibest];
 if (debug) printRZ(stderr, best,n);

 while (stack) {
  if (stack <(MAXSTACK/2)) {
     icheck = popr(rstack, &stack, MAX);
     if (icheck != EMPTY)
       rcheck = rstack[icheck];
    }
  else {
     icheck = popr(rstack, &stack, MIN);
     if (icheck != EMPTY)
       rcheck = rstack[icheck];
    }

  if (icheck != EMPTY) {
   if (better(rcheck, best)) {
    best = rcheck;
    if (debug) printRZ(stderr, best,n);
    pushAdd(&rcheck,n);
   } else if (coolvar(rcheck, best, coeff)) {
    pushAdd(&rcheck,n);
   }
  }
 }
 printRZ(stdout, best,n);

 
}



int main(int argc, char **argv) {

 char *p ;
 int i,n, multi = FALSE, isadd = TRUE, len = 0;

 debug = FALSE;

 /* nessun parametro: stampa help */
 if ((args[1] == NULL)) {
    fprintf(stdout,"usage: approz -(a|s)[m|d] <number> [<coeff>]\n\r");
    fprintf(stdout,"       approz \n\r");
    fprintf(stdout,"       approz -ver\n\r");
        
    fprintf(stdout," d(ebug) : stampa valori intermedi");
    fprintf(stdout," m(ultipla): ricerca sino a number"); 
    fprintf(stdout," null: ricerca il singolo numero");

    fprintf(stdout," a(dd): per aggiunte sucessive a partire da [0,1]");
    fprintf(stdout," s(ub): per selezione a partire da [0..n]");
    return FALSE;
  }

  /* stampa versione e copyright */
  if (!strcmp(args[1],"-ver")) {
    fprintf(stdout,"   Copyright (c) 2001, %s\n\r",AUTHOR);
    fprintf(stdout,"         All rights reserved.\n\r");
    fprintf(stdout,"   %s version %s\n\r",TITLE,VERSION);
    return FALSE;
  } else if

 /* primo parametro inizia per "-" ed e' lungo almeno 2 */  
 if ((argv[1][0] != '-') || (len = strlen(argv[1]) <2)){
  fprintf(stderr," Error: no command -(s|a)");
  exit(0);
 }

 /* primo comando */
 switch (argv[1][1]) {
  case 'a' : isadd = TRUE;
             break;
  case 's' : isadd = FALSE;
             break;
  default : fprintf(stderr," Error: unknow command");
            exit(0);
 }

 /* se c'e' il secondo comando lo applica */
 if (len > 2) {
  switch (argv[1][2]) {
  case 'm' : multi = TRUE;
             break;
  case 'd' : debug = TRUE;
             break;
  default : break;
  }
 }

 /* prende numero */
 if (argv[2] == NULL) {
   fprintf(stderr," Error: no number");
   exit(0);
 } else {
   n = atoi(argv[2]);
 }
 
 /* prende coefficente (opzionale) */
 if (argv[3] != NULL) {
   coeff = atof(argv[3]);
 } else {
   coeff = 0.5;
 } 

 /* multi scan/ numero singolo, add/sub */
 if (multi) {
  for (i=3; i<=n; i++) {
   
   if (isadd) 
   {
     addinit(i);
     addroz(i);
   } 
   else 
   {
     init(i);
     approz(i);
   }
  }

 } 
 else 
 {
  if (isadd) 
   {
     addinit(i);
     addroz(i);
   } 
  else 
   {
     init(i);
     approz(i);
   }
 }
}

}
