/*
 * Approssima i righelli ottimali di zanasi (RoZ)
 * facendo una ricerca euristica in profondita'
 *
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#define MAXTAG 10
  /* MAXN = [Tr(MAXTAG)] */
#define MAXN 55
#define MAXSTACK 3000
#define FALSE 0
#define TRUE 1
#define MIN 1
#define MAX 2


#define better(a,b) ((a.tac < b.tac) || (a.tac == b.tac) && (a.var < b.var))
#define notbad(a,best,level) ((a.tac - level) < best.tac)

typedef struct {

 int x[MAXN];      /* tacca i-esima */
 int chkroz[MAXN]; /* quante volte appare la distanza i */
 int tac;          /* numero tacche rz */
 int sum;          /* somma dei chkroz */
 float var;          /* varianza chkroz */
 float ave;          /* media chkroz */

} trighello;

trighello rstack[MAXSTACK];
int stack=0, debug = FALSE;

int equal(trighello *a, trighello *b) {
 int i;
 if (a->sum != b->sum) return FALSE;
 if (a->tac != b->tac) return FALSE;

 for (i=0; i<a->tac; i++) {
  if (a->x[i] != b->x[i]) return FALSE;
 }

 return TRUE;
}


int notinside(trighello *t, trighello *s, int stack) {
 int i;

 for (i=0; i<stack; i++){
  if (equal(t, s+i)) return FALSE;
 }

 return TRUE;
}

int gte(trighello *a, trighello *b) {
 int i;

 if (a->sum >= b->sum) return TRUE;
 if (a->tac >= b->tac) return TRUE;

 return FALSE;
}


int max(trighello *s, int stack) {
 int p = 0,i;

 for (i=0; i<stack; i++){
  if (gte(s+i, s+p)) p = i;
 }

 return p;
}

int min(trighello *s, int stack) {
 int p = 0,i;

 for (i=0; i<stack; i++){
  if (gte(s+p, s+i)) p = i;
 }

 return p;
}

/* mette l'elemento in pila solo se non gia' presente */
int pushr(trighello p, trighello *r, int *s) {
  if (*s<= MAXSTACK) {
   if ((*s == 0) || (notinside(&p, r, *s))) {
    r[*s] = p;
    (*s)++;
   }
  }
 else {
   fprintf(stderr, "Error: stack overflow\n");
   exit(1);
  }
}



/* restituisce un elemento */
/* se mode == MIN il migliore */
/* se mode == MAX il peggiore */
/* il primo in pila altrimenti */
trighello popr(trighello *r, int *s, int mode){
 int p;
 trighello swap;

 if (*s>0) {

  if (mode == MIN)
   p = min(r,stack);
  if (mode == MAX)
   p = max(r,stack);

  if ((mode == MIN) || (mode == MAX)){
   swap = r[p];
   r[p] = r[*s-1];
   r[*s-1] = swap;
  }

  return r[--(*s)];
 } else {
   fprintf(stderr, "Error: stack empty\n");
   exit(1);
  }
}

/* calcola il campo chkroz del righello
   calcola quante volte compare ogni possibile distanza*/
void calcChkroz(trighello *r, int lev){
 int i;

 if (lev < r->tac) {
  for (i=lev; i<r->tac; i++){
   r->chkroz[i-lev]++;
   r->sum++;
  }
  calcChkroz(r, lev+1);
 }

}

/* calcola sino a quanto conta il righello r */
int calcRZ(trighello *r, int n) {
 int i;

 for (i=-1;
     ((i+1<=n) && ( r->chkroz[i+1]));
     i++);


 return i;
}

/* calcola varianza semplificata ovvero average(abs(x-ave)) */
float varianza(trighello r) {
 float ave = r.ave;
 float sum = r.sum - r.x[0];
 float var = 0;
 int i;
 for (i = 1; i<r.tac; i++) {
   var += (float) r.x[i] * (float) abs((float)r.x[i] - ave);
 }

 var /= sum;

 return var;
}

float media(trighello r) {
 float ave;
 int i;

 ave = 0;

 for (i=0; i<r.tac; i++) {
  ave += r.x[i] * r.x[i] ;
 }

 if (r.sum)
   return ave/ (float)r.sum;
 else {
  fprintf(stderr, "Error: sum 0 on average");
  exit(1);
 }
}

/* la sottrazione [-] */
/* FIX: non ci sono controlli */
trighello sub(trighello *rinput, int p) {
 int i,dist;
 trighello r;
 float sum, var, ave;

 r = *rinput;


 /* ricalcola chkroz e sum */
 for (i=0; i<r.tac; i++){

  dist = abs(r.x[p]- r.x[i]);
  r.sum--;
  r.chkroz[dist]--;
 }

 /* sposta */
 for (i=p; i<r.tac; i++) {
  r.x[i] = r.x[i+1];
 }

 --r.tac;
 /* calcola media */
 r.ave = media(r);

 /* calcola varianza */
 r.var = varianza(r); 

 return r;
}

/* mette nello stack l'insieme degli r [-] a */
void pushSub(trighello *r, int a){
 trighello result;
 int i;

 for (i=0; i<r->tac; i++){
  result = sub(r, i);
  if (calcRZ(&result,a) == a)
   pushr(result, rstack, &stack) ;
 }
}

/* riempie lo stack cont tutti gli elementi
   dell'insieme [0...a] [-] xi */
void init(int a) {
 int i;
 trighello rinit, result;

 /* [0,1,..., a] */
 for (i=0; i<=a; i++){
   rinit.x[i] = i;
   rinit.chkroz[i]=0;
 }
 rinit.tac = a+1;
 rinit.sum = 0;

 calcChkroz(&rinit,0);

 /* calcola media */
 rinit.ave = media(rinit);

 /* calcola varianza */
 rinit.var = varianza(rinit);

 pushSub(&rinit, a);
}

void printRZ(FILE *f, trighello r, int n){
 int i;


 fprintf(f, "rz(%2u): (t=%2u s=%u) [", n, r.tac,stack);
 for (i=0; i<r.tac-1; i++){
  fprintf(f, "%2u,", r.x[i]);

 }
 fprintf(f, "%2u] (var = %3f ave = %3f)\n", r.x[r.tac-1] ,r.var, r.ave);
 fflush(f);
}

/*
 * estrae dalla pila
 * sottrai gli xi
 * rimette in coda se sono ancora
 */
void approz(int n){

 trighello rcheck,best;

 best = popr(rstack, &stack, MIN);
 if (debug) printRZ(stderr, best,n);

 while (stack) {
  if (stack < best.tac * best.tac )
    rcheck = popr(rstack, &stack, MAX);
  else
    rcheck = popr(rstack, &stack, MIN);

  if (better(rcheck, best)) {
   best = rcheck;
   if (debug) printRZ(stderr, best,n);
   pushSub(&rcheck,n);
  } else if (notbad(rcheck, best, 3)){
   pushSub(&rcheck,n);
  }
 }
 printRZ(stdout, best,n);
}

/* approz [m|d] <number> */
/* d --> debug */
/* m --> ricerca sino a n */
/* else --> ricerca il singolo numero */
int main(int argc, char **argv) {

 char *p ;
 int i,n,start=1, multi = FALSE;

 debug = FALSE;


 if (argv[1][0] == 'd') {
  debug = TRUE;
  start = 2;
 }
 if (argv[1][0] == 'm') {

  multi = TRUE;
  start = 2;
 }

 p = argv[start];
 if (p== NULL) {
   fprintf(stdout, "Error: no parameters\n");
   exit (0);
  }

 n = atoi(p);

 if (multi) {
  for (i=3; i<=n; i++) {
   init(i);
   approz(i);
  }

 } else {
  init(n);
  approz(n);
 }
}
